;;; org-contrib.el --- Unmaintained add-ons for Org-mode  -*- lexical-binding: t; -*-

;; Copyright (C) 2021 Bastien Guerry

;; Author: Bastien Guerry <bzg@gnu.org>
;; Homepage: https://git.sr.ht/~bzg/org-contrib
;; Package-Requires: ((emacs "25.1") (org "9.4.6"))
;; Version: 0.8
;; Keywords: org
;; SPDX-License-Identifier: GPL-3.0-or-later

;; This file is not part of GNU Emacs.

;; This program is free software: you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation, either version 3 of the License, or
;; (at your option) any later version.

;; This program is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with GNU Emacs.  If not, see <https://www.gnu.org/licenses/>.

;;; Commentary

;; This package contains add-ons to Org-mode which are not part of GNU
;; Emacs or of the official Org package.

;; ** This package receive little if no maintainance. **

;; Especially, there is no guaranty that it is compatible with the latest
;; Org stable version.  Would you would like to volunteer maintaining it?
;; If so, please send me an email at bzg@gnu.org.

;; If you want to maintain only one or some of these add-ons please send
;; me an email and consider hosting the add-ons on a separate repository.

;; These files used to live in the Org repository but have been filtered
;; out of the Org 9.5 release and are stored here for archival purpose.
;; The contrib/ directory used to contain a scripts/ directory that now
;; lives on Worg: <https://code.orgmode.org/bzg/worg/src/master/code>.

;;; Note:

;; This file, org-contrib.el, provides metadata for the (future)
;; org-contrib GNU ELPA package.

;;; Code:

(provide 'org-contrib)

;;; org-contrib.el ends here
